/* Auspice functions.
 *
 * Auspice Services is copyright (c) 2000-2001 In Mean.
 *     E-mail: <auspice@auspice.org>
 *
 * This program is free but copyrighted software; see the file LICENSE for
 * details.
 */

#include "../inc/services.h"
#include "../inc/timeout.h"

static BotInfo *makebot(const char *nick);
static BotInfo *botlists[256];
static void do_bot(const char *source);
static void alpha_insert_bot(BotInfo *bi);
static void change_bot_nick(BotInfo *bi, char *newnick);
static void insert_bot(BotInfo *bi);
static void do_help(const char *source);
static void do_ohelp(const char *source);
static void do_bot_settings(User *u, ChannelInfo *ci, const char *opts, const char *acmd);
static void do_info(const char *source);
static void do_chanset(const char *source);
static UserData *get_user_data(Channel *c, User *u);
void dropmode(User *u, Channel *c, ChannelInfo *ci);
void botkick(User *u2, ChannelInfo *ci, Channel *c, char *buf);
static void abotcom(const char *source, ChannelInfo *ci, char *buf, Channel *c);
static void doonu(User *u, ChannelInfo *ci, Channel *c, char *abot, char *areason);
/**********************************************************************/
void load_bs2_db(void)
{
    FILE *f = fopen(AUTH_DB, "r");
    int i, ver;
    BotInfo *bi;

    if (!(f = open_db(s_BotServ, BOTSERV2_DB, "r", 1)))
        return;

    switch (ver = get_file_version(f, BOTSERV2_DB)) {
	case 1:
        for (i = 33; i < 256; ++i) {
            while (fgetc(f) == 1) {
                   bi = smalloc(sizeof(BotInfo));
                   if (1 != fread(bi, sizeof(BotInfo), 1, f))
                       fatal_perror("Read error on %s", BOTSERV2_DB);

                if (bi->nick)
                    bi->nick = read_string(f, BOTSERV2_DB);
                alpha_insert_bot(bi);

                if (bi->user)
                    bi->user = read_string(f, BOTSERV2_DB);
                if (bi->host)
                    bi->host = read_string(f, BOTSERV2_DB);
                if (bi->real)
                    bi->real = read_string(f, BOTSERV2_DB);
                if (bi->master)
                    bi->master = read_string(f, BOTSERV2_DB);
		if (bi->flags == 0)
		    bi->flags |= BI_USER;
            } /* while (fgetc(f) == 1) */
        } /* for (i) */
	break;
      default:
        fatal("Unsupported version number (%d) on %s", i, BOTSERV2_DB);
    } /* switch (version) */

    close_db(f, BOTSERV2_DB);
}
/**********************************************************************/
static void alpha_insert_bot(BotInfo *bi)
{
    BotInfo *bi2, *bi3;
    char *nick = bi->nick;

    for (bi3 = NULL, bi2 = botlists[tolower(*nick)];
                        bi2 && stricmp(bi2->nick, nick) < 0;
                        bi3 = bi2, bi2 = bi2->next)
        ;
    bi->prev = bi3;
    bi->next = bi2;
    if (!bi3)
        botlists[tolower(*nick)] = bi;
    else
        bi3->next = bi;
    if (bi2)
        bi2->prev = bi;
}
/**********************************************************************/
void save_bs2_db(void)
{
    FILE *f;
    int i;
    BotInfo *bi;

    if (!(f = open_db(s_BotServ, BOTSERV2_DB, "w", BS_VERSION)))
        return;
    for (i = 33; i < 256; ++i) {
        for (bi = botlists[i]; bi; bi = bi->next) {
            fputc(1, f);

            if (1 != fwrite(bi, sizeof(BotInfo), 1, f))
            fatal_perror("Write error on %s", BOTSERV2_DB);
            if (bi->nick) {
                write_string(bi->nick, f, BOTSERV2_DB);
            }
            if (bi->user) {
                write_string(bi->user, f, BOTSERV2_DB);
            }
            if (bi->host) {
                write_string(bi->host, f, BOTSERV2_DB);
            }
            if (bi->real) {
                write_string(bi->real, f, BOTSERV2_DB);
            }
            if (bi->master) {
                write_string(bi->master, f, BOTSERV2_DB);
            }
        }
        fputc(0, f);
    }
    close_db(f, BOTSERV2_DB);
}
/**********************************************************************/
static BotInfo *makebot(const char *nick)
{
    BotInfo *bi;

    bi = scalloc(sizeof(BotInfo), 1);
    bi->nick = sstrdup(nick);
    alpha_insert_bot(bi);
    return bi;
}
/**********************************************************************/
static int delbot(BotInfo *bi)
{
    do_expireabot(bi->nick);
    if (bi->next)
        bi->next->prev = bi->prev;
    if (bi->prev)
        bi->prev->next = bi->next;
    else
        botlists[tolower(*bi->nick)] = bi->next;
    if (bi->nick)
      free(bi->nick);
    if (bi->user)
      free(bi->user);
    if (bi->host)
      free(bi->host);
    if (bi->real)
      free(bi->real);
    if (bi->master)
      free(bi->master);

    free(bi);
    return 1;
}
/**********************************************************************/
BotInfo *findbot(const char *nick)
{
    BotInfo *bi;
    if (!nick || !*nick) return NULL;
    for (bi = botlists[tolower(*nick)]; bi; bi = bi->next) {
        if (stricmp(bi->nick, nick) == 0)
            return bi;
    }
    return NULL;
}
/**********************************************************************/
void sbotserv(User *u, const char *source, char *buf)
{
    BotInfo *bi;
    char *cmd, *s, *chan, *bchan;
    cmd = strtok(buf, " ");

    if (cmd) {
            if (stricmp (cmd, "\1PING") == 0)
            {
                if (!(s = strtok (NULL, "")))
                    s = "\1";
                notice(s_BotServ, source, "\1PING %s", s);

            } else if (stricmp (cmd, "\1VERSION\1") == 0) {
                if (!(s = strtok (NULL, "")))
                    s = "\1";
                notice(s_BotServ, source, "%sVERSION Auspice IRC Services %s", s, s);
            } else if (stricmp(cmd, "HELP") == 0) {
			do_help(source);
            } else if (stricmp(cmd, "MUTE") == 0) {
                int i;
		if (!is_services_coder(source))
			return;
		for (i = 33; i < 256; ++i)
		for (bi = botlists[i]; bi; bi = bi->next)
		send_cmd(s_NickServ, "SVSMODE %s +d 0", bi->nick);
            } else if (stricmp(cmd, "UNMUTE") == 0) {
                int i;
		if (!is_services_coder(source))
			return;
		for (i = 33; i < 256; ++i)
		for (bi = botlists[i]; bi; bi = bi->next)
		send_cmd(s_NickServ, "SVSMODE %s -d 0", bi->nick);
            } else if (stricmp(cmd, "LIST") == 0) {
                int i;
                notice(s_BotServ, source, "\2Bot Listing\2.");
		for (i = 33; i < 256; ++i)
			for (bi = botlists[i]; bi; bi = bi->next)
			if (bi->flags & BI_USER) {
				notice(s_BotServ, source, "%-10s \2%-15s\2 (%s@%s) - %s", "Public",bi->nick, bi->user, bi->host, bi->master);
			} else if (is_uoper(u)) {
				notice(s_BotServ, source, "%-10s \2%-15s\2 (%s@%s) - %s", bi->flags & BI_MASTER?"Private":"NotSet",bi->nick, bi->user, bi->host, bi->master);
			}
                notice(s_BotServ, source, "\2End OF List\2.");

            } else if (stricmp(cmd, "SET") == 0) {
		ChannelInfo *ci;
		char *opts, *acmd;
		User *u = finduser(source);

		if (readonly) {
		    notice(s_BotServ, source, 
                	"Sorry, channel option setting is temporarily disabled.");
	            return;
	        }
        
               chan = strtok(NULL, " ");
               opts = strtok(NULL, " ");
               acmd = strtok(NULL, " ");

	       if (!chan) {
		    notice(s_BotServ, source, 
                	"Syntax: /Msg %s SET #channel options ON|OFF", s_BotServ);
		    return;
	       }

	       ci = cs_findchan(chan);

	       if (!ci) {
		    notice(s_BotServ, source, 
                	"Channel %s is not registered.", chan);
		    return;
	       }

	       if (!u) {
			return;
	       }

	       if ((get_access(u,ci) < 12) && !is_oper_cando(source, 16)) {
		    notice(s_BotServ, source, 
                	"Only Founder or Co Founder can change bot setting.");
		    return;
	       }

	       if (!opts || !acmd) {
			notice(s_BotServ, u->nick, "Syntax: /Msg %s SET #channel <Option> <ON|OFF>", s_BotServ);
	       } else if (stricmp(acmd,"ON") && stricmp(acmd,"OFF")) {
			notice(s_BotServ, u->nick, "Syntax: /Msg %s SET #channel <Option> <ON|OFF>", s_BotServ);
	       } else {	
		       do_bot_settings(u,ci,opts,acmd);
	       }

            } else if (stricmp(cmd, "ASSIGN") == 0) {
		ChannelInfo *ci;
		User *u = finduser(source);

		if (readonly) {
		    notice(s_BotServ, source, 
                	"Sorry, channel option setting is temporarily disabled.");
	            return;
	        }
        
                chan = strtok(NULL, " ");
                bchan = strtok(NULL, " ");

	       if (!chan || !bchan) {
		    notice(s_BotServ, source, 
                	"Syntax: /Msg %s ASSIGN #channel botnick", s_BotServ);
		    return;
	       }

	       ci = cs_findchan(chan);

	       if (!ci) {
		    notice(s_BotServ, source, 
                	"Channel %s is not registered.", chan);
		    return;
	       }

	       if (!u) return;
	      
	       if ((get_access(u,ci) < 12) && !is_oper_cando(source, 16)) {
		    notice(s_BotServ, source, 
                	"Only Founder or Co Founder can assign bot.");
		    return;
	       }

	       bi = findbot(bchan);

	       if (!bi) {
		    notice(s_BotServ, source, 
                	"This bot is not exist, please use command /Msg %s LIST   to list available bot.", s_BotServ);
		    return;
	       }

	      if (stricmp(source, bi->master) && (bi->flags & BI_MASTER) && !is_oper_cando(source, 1)) {
		    notice(s_BotServ, source, 
                	"Only Bot Master can assign this bot");
		    return;
	       }

	       if ((ci->bot) && (!stricmp(ci->bot, bchan))) {
		    notice(s_BotServ, source, "This bot already assigned");
		    return;
	       }
               csbot++;
               if(ci->bot) {
                        botpart3(chan);
                        free(ci->bot);
               }
               ci->bot = sstrdup (bchan);
               botjoin(chan);
               notice(s_BotServ, source, "Your channel bot is \2%s\2.", bchan);
		if (!(ci->botflag & CBI_BOTCOM)) {
			ci->botflag |= CBI_BOTCOM;
			notice(s_BotServ, source, "BOTCOM options is automaticly enable for your channel, to disable it use command: /msg BotServ SET %s BOTCOM OFF", chan);
                }
               slog("CS BOT %s bot: %s", chan, ci->bot);
               show_u_next_db(u, s_BotServ);
            } else if (stricmp(cmd, "UNASSIGN") == 0) {
		ChannelInfo *ci;
		User *u = finduser(source);

		if (readonly) {
		    notice(s_BotServ, source, 
                	"Sorry, channel option setting is temporarily disabled.");
	            return;
	        }
        
                chan = strtok(NULL, " ");

	       if (!chan) {
		    notice(s_BotServ, source, 
                	"Syntax: /Msg %s UNASSIGN #channel", s_BotServ);
		    return;
	       }

	       ci = cs_findchan(chan);

	       if (!ci) {
		    notice(s_BotServ, source, 
                	"Channel %s is not registered.", chan);
		    return;
	       }

	       if (!u) {
			return;
	       }

	       if ((get_access(u,ci) < 12) && !is_oper_cando(source, 16)) {
		    notice(s_BotServ, source, 
                	"Only Founder or Co Founder can unassign bot.");
		    return;
	       }

	       if (!ci->bot) {
		    notice(s_BotServ, source, 
                	"There is not bot in this channel");
		    return;
	       }
               csbot++;
               if(ci->bot) {
                        botpart3(chan);
                        free(ci->bot);
                        ci->bot = NULL;
               }
	       notice(s_BotServ, source, "There are no bot set in your channel now.");
               slog("CS BOT %s unset bot", chan);
               show_u_next_db(u, s_BotServ);
            } else if (stricmp(cmd, "BOT") == 0) {
		do_bot(source);
            } else if (stricmp(cmd, "INFO") == 0) {
		do_info(source);
            } else if (stricmp(cmd, "CHANSET") == 0) {
		do_chanset(source);
            } else if (stricmp(cmd, "WHOIS") == 0) {
		struct tm tm;
		char timebuf[64];
                chan = strtok(NULL, " ");
                if (!chan) {
                        notice(s_BotServ, source, "Syntax: WHOIS \2Botnick\2");
                } else {
			bi = findbot(chan);
                        if (bi) {
			        char bufss[512];
				*timebuf =0;
				notice(s_BotServ, source,"Bot Nick   : %s (%s@%s) %s", bi->nick, bi->user, bi->host, bi->real);
				notice(s_BotServ, source,"Master     : %s", bi->master);
			        tm = *localtime(&bi->created);
			        strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);
			        timebuf[sizeof(timebuf)-1] = 0;
			        notice(s_BotServ, source,"Registered : %s", timebuf);
				*bufss = 0;
				if (bi->flags & BI_MASTER)
				strcat(bufss, "Private");
				if (bi->flags & BI_USER) {
			        if (*bufss)
					strcat(bufss, ", ");
				strcat(bufss, "Public");
				}
			        notice(s_BotServ, source,"Options: %s", bufss);
                        }
                }

            } else if (stricmp(cmd, "OHELP") == 0) {
			do_ohelp(source);

            } else if (stricmp(cmd, "SNOOP") == 0) {
                if (!is_services_root(source)) {
                        notice(s_BotServ, source, PMD);
                        return;
                }

                chan = strtok(NULL, " ");
                bchan = strtok(NULL, " ");

		if (!chan || !bchan) {
			notice(s_BotServ, source, "Syntax: SNOOP [\2source\2] [\2dest\2]");
                        notice(s_BotServ, source, "Disable Channel Spy feature");
                        aspychan = 0;
                } else {
                        if (asource) free(asource);
                        asource = sstrdup(chan);
                        if (adest) free(adest);
                        adest = sstrdup(bchan);
                        aspychan = 1;
                        notice(s_BotServ, source, "All message in channel %s forward to %s.", chan, bchan);
                }

	    }
	}
}
 
/**************************************************************/
static void do_info(const char *source)
{
	ChannelInfo *ci;
	char *chan = strtok(NULL, " ");

	if (!chan) {
	        notice(s_BotServ, source, "Syntax: /MSG %s \2INFO\2 #<channel>", s_BotServ);
	} else {
		ci = cs_findchan(chan);
		if (!ci) {
		        notice(s_BotServ, source, "This channel is not a registered channel");
		} else if (!ci->bot) {
			notice(s_BotServ, source, "There is not bot assigned in this channel");
		} else {
        	    notice(s_BotServ, source, "Channel Bot : \2%s\2",ci->bot);
        	    notice(s_BotServ, source, "Caps Limit  : Allow: \2%d\2 Percent: \2%d\2", ci->capsmin, ci->capspercent);
        	    notice(s_BotServ, source, "Flood Limit : Lines: \2%d\2 Seconds: \2%d\2", ci->floodlines, ci->floodsecs);
        	    notice(s_BotServ, source, "Repeat Limit: Lines: \2%d\2", ci->repeattimes);
	            if (ci->botflag) {
        	        char buf[512];
                	*buf = 0;
		        if (ci->botflag & CBI_BOTCOM)
                	    strcat(buf, "Bot Command");
	                if (ci->botflag & CBI_BOTSTAY) {
        	            if (*buf)
	                        strcat(buf, ", ");
        	            strcat(buf, "Stay");
	                }
        	        if (ci->botflag & CBI_KICKCAP) {
                	    if (*buf)
                	        strcat(buf, ", ");
	                    strcat(buf, "Caps");
        	        }
	                if (ci->botflag & CBI_KICKBOLDS) {
        	            if (*buf)
                	        strcat(buf, ", ");
	                    strcat(buf, "Bolds");
        	        }
                	if (ci->botflag & CBI_KICKCOLORS) {
	                    if (*buf)
        	                strcat(buf, ", ");
                	    strcat(buf, "Colors");
	                }
        	        if (ci->botflag & CBI_KICKREVERSES) {
	                    if (*buf)
        	                strcat(buf, ", ");
                	    strcat(buf, "Reverses");
	                }
        	        if (ci->botflag & CBI_KICKUNDERLINES) {
                	    if (*buf)
                        	strcat(buf, ", ");
	                    strcat(buf, "Underlines");
        	        }
	                if (ci->botflag & CBI_KICKBADWORDS) {
        	            if (*buf)
                	        strcat(buf, ", ");
	                    strcat(buf, "Badwords");
        	        }
	                if (ci->botflag & CBI_KICKFLOOD) {
        	            if (*buf)
                	        strcat(buf, ", ");
	                    strcat(buf, "Flood");
        	        }
	                if (ci->botflag & CBI_KICKREPEAT) {
        	            if (*buf)
                	        strcat(buf, ", ");
	                    strcat(buf, "Repeat");
        	        }
                	if (ci->botflag & CBI_PROTOP) {
	                    if (*buf)
        	                strcat(buf, ", ");
	                    strcat(buf, "Protect Op");
        	        }
	               notice(s_BotServ, source, "Bot Options : \2%s\2", buf);
	            } /* if have flag */
		} /* !ci? */
	} /* !chan? */
}
/*************************************************************/
static void do_chanset(const char *source)
{
	ChannelInfo *ci;
	char *chan = strtok(NULL, " ");
	char *cmd = strtok(NULL, " ");
	char *param1 = strtok(NULL, " ");
	char *param2 = strtok(NULL, " ");

	if (!chan || !cmd || !param1) {
	        notice(s_BotServ, source, "Syntax: /MSG %s \2CHANSET\2 #<channel> <option> <param1> <param2>", s_BotServ);
	} else {
		ci = cs_findchan(chan);
		if (!ci) {
		        notice(s_BotServ, source, "This channel is not a registered channel");
		} else if (!ci->bot) {
			notice(s_BotServ, source, "There is not bot assigned in this channel");
		} else {
	            User *u = finduser(source);
               if (!u) {
                        return;
               }

               if ((get_access(u,ci) < 12) && !is_oper_cando(source, 16)) 
		{
                    notice(s_BotServ, source,
                        "Only Founder or Co Founder can change bot setting.");
                    return;
               }

		    if (!stricmp(cmd, "CAPS")) {
			if (!param2) {
			       notice(s_BotServ, source, "Syntax: /MSG %s \2CHANSET\2 %s CAPS <allow> <percents>", s_BotServ, chan);
			} else if ((strlen(param1) > 3) || (strlen(param2) > 2)) {
				notice(s_BotServ, source, "Input too big");
			} else {
				ci->capsmin = atoi(param1);
				ci->capspercent = atoi(param2);
		       	    notice(s_BotServ, source, "Caps Limit  : Allow: \2%d\2 Percent: \2%d\2", ci->capsmin, ci->capspercent);
			}
		    } else if (!stricmp(cmd, "FLOOD")) {
			if (!param2) {
			       notice(s_BotServ, source, "Syntax: /MSG %s \2CHANSET\2 %s FLOOD <lines> <secs>", s_BotServ, chan);
			} else if ((strlen(param1) > 3) && (strlen(param2) > 2)) {
				notice(s_BotServ, source, "Input too big");
			} else {
				ci->floodlines = atoi(param1);
				ci->floodsecs = atoi(param2);
	        	    	notice(s_BotServ, source, "Flood Limit : Lines: \2%d\2 Seconds: \2%d\2", ci->floodlines, ci->floodsecs);
			}
		    } else if (!stricmp(cmd, "REPEAT")) {
			if ((strlen(param1) > 3)) {
				notice(s_BotServ, source, "Input too big");
			} else {
				ci->repeattimes = atoi(param1);
		       	    notice(s_BotServ, source, "Repeat Limit: Lines: \2%d\2", ci->repeattimes);
			}
		    } else {
		       	    notice(s_BotServ, source, "Unknown Option.");
		    }
		} /* !ci? */
	} /* !chan? */
}
/*************************************************************/
static void do_bot(const char *source)
{
    BotInfo *bi;
    NickInfo *ni = findnick(source);
    char *cmd = strtok(NULL, " ");
    if (!cmd) {
        notice(s_BotServ, source, "Syntax: /MSG %s \2BOT\2 <command> [<parameters>].", s_BotServ);
    } else if (!is_oper_cando(source, 16) || !ni) {
        notice(s_BotServ, source, PMD);
    } else if (!stricmp(cmd, "ADD")) {
        char *nick = strtok(NULL, " ");
        char *user = strtok(NULL, " ");
        char *host = strtok(NULL, " ");
        char *real = strtok(NULL, "");

        if (!nick || !user || !host || !real)  {
		notice(s_BotServ, source,"Syntax: BOT ADD \2botnick\2 \2username\2 \2bothost\2 \2realname\2");
		notice(s_BotServ, source, "Example: BOT ADD testbot MyNewBot botserv.auspice.org In Mean Bot");
        } else if (findbot(nick)) {
                notice(s_BotServ, source, "This bot already exist.");
	} else {
		if (finduser(nick)) {
			notice(s_BotServ, source, "This nick currently online.");
			return;
		}
                bi = makebot(nick);
                if (!bi) {
	                notice(s_BotServ, source, "Failed created bot.");
			return;
                }
                bi->user = sstrdup(user);
                bi->host = sstrdup(host);
                bi->real = sstrdup(real);
                bi->master = sstrdup(source);
                bi->created = CTime;
		bi->flags = 0;
		bi->flags |= BI_USER;
                bi->chancount = 0;
                NEWNICK(bi->nick, bi->user, bi->host, bi->real, BOTMODE, 1);
                notice(s_BotServ, source, "Bot \2%s\2 connected.", bi->nick);
	}
     } else if (!stricmp(cmd, "CHANGE")) {
        char *oldnick = strtok(NULL, " ");
        char *nick = strtok(NULL, " ");
        char *user = strtok(NULL, " ");
        char *host = strtok(NULL, " ");
        char *real = strtok(NULL, "");
	int is_sr = is_services_root(source);

        if (!oldnick || !nick)
                notice(s_BotServ, source, "Syntax: BOT CHANGE \2old-bot-nick\2 \2new-bot-nick\2 [\2user\2] [\2host\2] [\2real\2]");
        else if (!(bi = findbot(oldnick)))
                notice(s_BotServ, source, "Bot %s doesn't exist", oldnick);
	else if (!is_sr && (stricmp(source, bi->master) != 0)) {
	        notice(s_BotServ, source, PMD);
		return;
        } else {
                if (!strcmp(bi->nick, nick)
                                && ((user) ? !strcmp(bi->user, user) : 1)
                                && ((host) ? !strcmp(bi->host, host) : 1)
                                && ((real) ? !strcmp(bi->real, real) : 1))
                {
                        notice(s_BotServ, source, "Setting didn't changed.");
                        return;
                }
                if (stricmp(bi->nick, nick) && findbot(nick))
                {
                        notice(s_BotServ, source, "Bot %s already existed", nick);
                        return;
                }
                if (stricmp(bi->nick, nick) && finduser(nick))
                {
                        notice(s_BotServ, source, "nick %s currently online", nick);
                        return;
                }
                if (stricmp(bi->nick, nick))
                {
			User *u2;
                        send_cmd(NULL, "UNSQLINE %s", bi->nick);
                        if ((u2 = finduser(nick)))
                                kill_user(s_BotServ, u2->nick, "This nick is now used by Services");
                }
                if (strcmp(nick, bi->nick))
                        change_bot_nick(bi, nick);
                if (user && strcmp(user, bi->user))
                {
                        free(bi->user);
                        bi->user = sstrdup(user);
                }
                if (host && strcmp(host, bi->host))
                {
                        free(bi->host);
                        bi->host = sstrdup(host);
                }
                if (real && strcmp(real, bi->real))
                {
                        free(bi->real);
                        bi->real = sstrdup(real);
                }
                if (!user)
                        send_cmd(oldnick, "NICK %s", bi->nick);
                else
                {
			do_checkbotpart(oldnick, bi->nick);
                        send_cmd(oldnick, "QUIT :Quit: Be right back");
	                NEWNICK(bi->nick, bi->user, bi->host, bi->real, BOTMODE, 1);
                        do_checkbot2(bi->nick);
                }
                notice(s_BotServ, source, "Bot \2%s\2 has been changed to %s!%s@%s (%s)", oldnick, bi->nick, bi->user, bi->host, bi->real);
	}
     } else if (!stricmp(cmd, "TRANSFER")) {
        char *nick = strtok(NULL, " ");
        char *master = strtok(NULL, " ");
	int is_sr = is_services_root(source);

        if (!nick || !master)
                notice(s_BotServ, source, "Syntax: BOT TRANSFER \2bot-nick\2 \2new-master-nick\2");
        else if (!(bi = findbot(nick)))
                notice(s_BotServ, source, "Bot %s doesn't exist", nick);
        else if (!findnick(master))
                notice(s_BotServ, source, "New master nick is not registered");
	else if (!is_sr && (stricmp(source, bi->master))) {
	        notice(s_BotServ, source, PMD);
		return;
        } else {
                if (stricmp(master, bi->master))
                {
                        free(bi->master);
                        bi->master = sstrdup(master);
                }
                notice(s_BotServ, source, "Bot \2%s\2 has been changed master to (%s)", nick, bi->master);
	}
     } else if (!stricmp(cmd, "SET")) {
        char *nick = strtok(NULL, " ");
        char *opts = strtok(NULL, " ");
        char *choo = strtok(NULL, " ");
	int is_sr = is_services_root(source);

	if (!nick || !opts || !choo) {
		notice(s_BotServ, source,"Syntax: BOT SET \2botnick\2 \2option\2 \2ON|OFF\2");
		return;
	}
	bi = findbot(nick);
	if (!bi) {
		notice(s_BotServ, source,"This services bot does not exist.");
		return;
	}
	if (stricmp(opts, "MASTER")==0) {
		if (!is_sr && stricmp(source, bi->master)) {
		        notice(s_BotServ, source, PMD);
			return;
		}
		if (stricmp(choo, "ON")==0) {
			bi->flags |= BI_MASTER;
			bi->flags &= ~BI_USER;
			notice(s_BotServ, source,"This services bot can assigned or use by the master only.");
		} else if (stricmp(choo, "OFF")==0) {
			bi->flags |= BI_USER;
			bi->flags &= ~BI_MASTER;
			notice(s_BotServ, source,"This services bot is public for everyone.");
		} else {
			notice(s_BotServ, source,"Syntax: BOT SET \2botnick\2 \2MASTER\2 \2ON|OFF\2.");
		}	
	} else {
		notice(s_BotServ, source,"Syntax: BOT SET \2botnick\2 \2option\2 \2ON|OFF\2");
	}
     } else if (!stricmp(cmd, "DEL")) {
        char *nick = strtok(NULL, " ");

        if (!nick)
		notice(s_BotServ, source,"Syntax: BOT DEL \2botnick\2");
        else if (!(bi = findbot(nick)))
                notice(s_BotServ, source, "This bot does not existed.");
        else
        {
                send_cmd(bi->nick, "QUIT :Quit: Deleted by %s!", source);
                send_cmd(NULL, "UNSQLINE %s", bi->nick);
                delbot(bi);
		notice(s_BotServ, source,"Deleted bot \2%s\2.", nick);
	}
    }
}
/**************************************************************/
void connectbot2()
{
	int i;
	BotInfo *bi;
	for (i = 33; i < 256; ++i)
		for (bi = botlists[i]; bi; bi = bi->next)
		if (bi) {
	                NEWNICK(bi->nick, bi->user, bi->host, bi->real, BOTMODE, 1);
		}
}
/**************************************************************/
void allbotjoin(const char *chan)
{
	int i;
	BotInfo *bi;
	for (i = 33; i < 256; ++i)
		for (bi = botlists[i]; bi; bi = bi->next)
		if (bi) {
                      send_cmd(bi->nick, "JOIN %s", chan);
		}
}
void allbotpart(const char *chan)
{
	int i;
	BotInfo *bi;
	for (i = 33; i < 256; ++i)
		for (bi = botlists[i]; bi; bi = bi->next)
		if (bi) {
                      send_cmd(bi->nick, "PART %s :I'm quit.", chan);
		}
}
/*************************************************************/
int check_ifbot(const char *abotname)
{
	if (findbot(abotname))
           return 1;
        return 0;
}
/************************************************************/
static void change_bot_nick(BotInfo *bi, char *newnick)
{
        if (bi->next)
        bi->next->prev = bi->prev;
    if (bi->prev)
        bi->prev->next = bi->next;
    else
    botlists[tolower(*bi->nick)] = bi->next;

        if (bi->nick) free(bi->nick);
        bi->nick = sstrdup(newnick);

        insert_bot(bi);
}
/***************************************/
static void insert_bot(BotInfo *bi)
{
    /* bi->prev = NULL;
    bi->next = botlist;
    if (bi->next) bi->next->prev = bi;
    botlist = bi; */

    BotInfo *ptr, *prev;

    for (prev = NULL, ptr = botlists[tolower(*bi->nick)];
                        ptr != NULL && stricmp(ptr->nick, bi->nick) < 0;
                        prev = ptr, ptr = ptr->next)
        ;
        bi->prev = prev;
    bi->next = ptr;
    if (!prev)
        botlists[tolower(*bi->nick)] = bi;
    else
        prev->next = bi;
    if (ptr)
        ptr->prev = bi;
}
/************************************************/
void botjoin2(ChannelInfo *ci)
{
        BotInfo *bi;
        Channel *c;
	if (!ci)
		return;
	c = findchan(ci->name);
	if (c && (c->havebot == 1))
		return;

        if (ci && ci->bot) {
		bi = findbot(ci->bot);
                if (bi) {
                      send_cmd(ci->bot, "JOIN %s", ci->name);

#if defined(PROTECTA)
                      send_cmd(ci->bot, "MODE %s +oa %s %s", ci->name, ci->bot, ci->bot);
#elif defined(LIQUID)
                      send_cmd(ci->bot, "MODE %s +ou %s %s", ci->name, ci->bot, ci->bot);
#else
                      send_cmd(ci->bot, "MODE %s +o %s", ci->name, ci->bot);
#endif

                      if (c)
                                c->havebot = 1;
                } else {
                        free(ci->bot);
			ci->bot = NULL;
                }
        }

}
/************************************************/
void botpart2(ChannelInfo *ci, char *newbot)
{
        Channel *c = findchan(ci->name);
        send_cmd(ci->bot, "PART %s :Be right back", ci->name);
	if (c) 
		c->havebot = 0;
	if (ci->bot) {
		free(ci->bot);
		ci->bot = sstrdup(newbot);
	}

}
/************************************************/
void bsconvert(char *nick, char *master)
{
	BotInfo *bi;
        bi = makebot(nick);
        if (!bi)
		return;
        bi->user = "BotServ";
        bi->host = sstrdup(services_host);
        bi->real = "Services Bot";
        bi->master = sstrdup(master);
        bi->created = CTime;
	bi->flags = 0;
	bi->chancount = 0;
}
/************************************************/
int is_botowner(const char *nick, const char *buf)
{
	BotInfo *bi = findbot(buf);
	if (nick && buf && bi)
		if (stricmp(bi->master, nick) == 0)
			return 1;
	return 0;
}
/*************************************************/
static void do_help(const char *source)
{
    const char *cmd = strtok(NULL, "");
    char buf[256];
    snprintf(buf, sizeof(buf), "%s%s", s_BotServ, cmd ? " " : "");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_BotServ, source, buf);
    if (is_oper_cando(source, 16)) {
        notice(s_BotServ, source, "For more Help using command /Msg %s HELP OHELP.", s_BotServ);
    }
}
static void do_ohelp(const char *source)
{
    const char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];
    snprintf(buf, sizeof(buf), "%s%s", s_BotServ, cmd ? " oper " : " oper");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_BotServ, source, buf);
    if (!cmd)
        slog("BS - %s requested ohelp: (index)", source);

    if (cmd)
        slog("BS - %s requested ohelp: %s", source, cmd);
}

/************************************************/
static void do_bot_settings(User *u, ChannelInfo *ci, const char *opts, const char *acmd) {

    if (!stricmp(opts, "BOTCOM")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_BOTCOM;
	        notice(s_BotServ, u->nick, "Active bot command in channel.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_BOTCOM;
	        notice(s_BotServ, u->nick, "Disable bot command in channel.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "KICKCAP")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_KICKCAP;
	        notice(s_BotServ, u->nick, "Bot will watch all setence for excess CAPs and ban kick.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_KICKCAP;
	        notice(s_BotServ, u->nick, "Disable Caplock control.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "BOTSTAY")) {
	if (!is_oper_cando(u->nick, 16)) {
	        notice(s_BotServ, u->nick, ERR_MOREFLAG);
		return;
	}
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_BOTSTAY;
	        notice(s_BotServ, u->nick, "Bot will now stay in the channel.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_BOTSTAY;
	        notice(s_BotServ, u->nick, "Bot will stay in the channel with atleast one user.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "BOLDS")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_KICKBOLDS;
	        notice(s_BotServ, u->nick, "Bot will now kick \2Bolds\2.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_KICKBOLDS;
	        notice(s_BotServ, u->nick, "Bot won't kick \2Bolds\2 anymore");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "COLORS")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_KICKCOLORS;
	        notice(s_BotServ, u->nick, "Bot will now kick \2Colors\2.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_KICKCOLORS;
	        notice(s_BotServ, u->nick, "Bot won't kick \2Colors\2 anymore.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "GREET")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_GREET;
	        notice(s_BotServ, u->nick, "Bot will now greet the user info.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_GREET;
	        notice(s_BotServ, u->nick, "Bot will not greet user on join.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "REVERSES")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_KICKREVERSES;
	        notice(s_BotServ, u->nick, "Bot will now kick \2Reverses\2.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_KICKREVERSES;
	        notice(s_BotServ, u->nick, "Bot won't kick \2Reverses\2 user.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "UNDERLINES")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_KICKUNDERLINES;
	        notice(s_BotServ, u->nick, "Bot will now kick \2Underlines\2.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_KICKUNDERLINES;
	        notice(s_BotServ, u->nick, "Bot won't kick \2Underlines\2 anymore.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "BADWORDS")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_KICKBADWORDS;
	        notice(s_BotServ, u->nick, "Bot will now kick \2Badwords\2.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_KICKBADWORDS;
	        notice(s_BotServ, u->nick, "Bot won't kick \2Badwords\2 anymore.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "REPEAT")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_KICKREPEAT;
	        notice(s_BotServ, u->nick, "Bot will now kick \2repeat\2.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_KICKREPEAT;
	        notice(s_BotServ, u->nick, "Bot won't kick \2repeat\2 anymore.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "FLOOD")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_KICKFLOOD;
	        notice(s_BotServ, u->nick, "Bot will now kick \2flood\2.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_KICKFLOOD;
	        notice(s_BotServ, u->nick, "Bot won't kick \2flood\2 anymore.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "PROTECTOP")) {
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_PROTOP;
	        notice(s_BotServ, u->nick, "Bot will now \2Protect Op\2.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_PROTOP;
	        notice(s_BotServ, u->nick, "Bot won't \2Protect Op\2  anymore.");
	        show_u_next_db(u, s_BotServ);
	}
#ifdef GAMES
    } else if (!stricmp(opts, "GUESSNUM") && is_tokn(39)) {
        if (!is_oper_cando(u->nick, 16)) {
                notice(s_BotServ, u->nick, ERR_MOREFLAG);
                return;
        }
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_GUESSNUM;
	        notice(s_BotServ, u->nick, "Activate Guessing Number Game in your channel.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_GUESSNUM;
	        notice(s_BotServ, u->nick, "DeActivate Guessing Number Game.");
	        show_u_next_db(u, s_BotServ);
	}
    } else if (!stricmp(opts, "GUESSWRM") && is_tokn(39)) {
        if (!is_oper_cando(u->nick, 16)) {
                notice(s_BotServ, u->nick, ERR_MOREFLAG);
                return;
        }
        if (!stricmp(acmd, "ON")) {
	        ci->botflag |= CBI_GUESSWRD;
	        notice(s_BotServ, u->nick, "Activate Guessing Word Game in your channel.");
	        show_u_next_db(u, s_BotServ);
	} else {
	        ci->botflag &= ~CBI_GUESSWRD;
	        notice(s_BotServ, u->nick, "DeActivate Guessing Word Game.");
	        show_u_next_db(u, s_BotServ);
	}
#endif
    } else {
	notice(s_BotServ, u->nick, "This option is not exist, please using command /Msg %s HELP SET for more info", s_BotServ);
    }
}
static UserData *get_user_data(Channel *c, User *u)
{
        struct c_userlist *user;
        if (!c || !u) return NULL;

        for (user = c->users; user; user = user->next) {
                if (user->user == u) {
                        if (user->ud)
                        {
                                time_t now = CTime;
                                /* Checks whether data is obsolete */
                                if (now - user->ud->last_use > 600)
                                {
					        if (user->ud->lastline) free(user->ud->lastline);
		        			/* We should not free and realloc, but reset to 0
			                           instead. */
					        memset(user->ud, 0, sizeof(UserData));
                                	        user->ud->last_use = now;
                                }

                                return user->ud;
                        } else {
                                user->ud = scalloc(sizeof(UserData), 1);
                                user->ud->last_use = CTime;
                                return user->ud;
                        }
                }
        }

        return NULL;
}
void botjoin(char *chan)
{
	ChannelInfo *ci = cs_findchan(chan);
	botjoin2(ci);
}
/*******************************************************************/
void botpart(Channel *c)
{
	ChannelInfo *ci;
	ci = cs_findchan(c->name);
	if (ci && ci->bot && !(ci->botflag & CBI_BOTSTAY)) {
	        send_cmd(ci->bot, "%s %s", istoken?TOK_PART:MSG_PART, ci->name);
		c->havebot = 0;
	}
}

void botpart3(char *chan)
{
	ChannelInfo *ci;
	Channel *c = findchan(chan);
	ci = cs_findchan(chan);
	if (ci && ci->bot)
	        send_cmd(ci->bot, "%s %s", istoken?TOK_PART:MSG_PART, ci->name);	
	if (c)
		c->havebot = 0;
}
void botpart4(char *chan)
{
	ChannelInfo *ci;
	Channel *c = findchan(chan);
	ci = cs_findchan(chan);
	if (ci && ci->bot)
	        send_cmd(ci->bot, "%s %s", istoken?TOK_PART:MSG_PART, ci->name);	
	if (c) {
		c->havebot = 0;
		change_cmode (s_ChanServ, chan, "-r", "");
	}
}
/*******************************************************************/
void botserv(const char *source, const char *thebot, char *buf)
{
    char *cmd, *s, *chan, *bmsg;
    char orig[BUFSIZE];
    NickInfo *ni;
    BotInfo *bi;
    strscpy (orig, buf, sizeof (orig));
    cmd = strtok(buf, " ");

    if (cmd) {
            if (stricmp (cmd, "\1PING") == 0) {
                if (!(s = strtok (NULL, "")))
                    s = "\1";
                notice(thebot, source, "\1PING %s", s);

            } else if (stricmp (cmd, "\1VERSION\1") == 0) {
                if (!(s = strtok (NULL, "")))
                    s = "\1";
             notice(thebot, source, "%sVERSION Auspice IRC Services %s", s, s);
	
           } else if (!is_oper_cando(source, 16)) {
        	chan = strtok(NULL, "");
		bi = findbot(thebot);
		if (bi) {
		    	ni = findnick(bi->master);

		    	if (ni && finduser(bi->master))
		            	if ((ni->flags & NI_IDENTIFIED) && (ni->flags & NI_BOTMSG))
				     	send_cmd(thebot, "%s %s :(%s) %s %s", istoken?TOK_PRIVATE:MSG_PRIVATE, ni->nick, source, cmd, (chan)?chan:"");
		}
	    	return;
          
            } else if (stricmp(cmd, "JOIN") == 0) {
                chan = strtok(NULL, " ");
                if (!chan) {
                        notice(thebot, source, "Syntax: JOIN \2Channel\2");
                } else {
			if (is_botowner(source, thebot) || is_services_root(source))
	                        send_cmd (thebot, "%s %s", istoken?TOK_JOIN:MSG_JOIN, chan);
			else
	                        notice(thebot, source, "Sorry you are not my master.");		
                }
            } else if (stricmp(cmd, "HELP") == 0) {
		if (is_botowner(source, thebot) || is_services_root(source)) {
		notice(thebot, source, "- JOIN      Join channel");	
		notice(thebot, source, "- PART      Part channel");	
		notice(thebot, source, "- SAY       Say in channel");	
		notice(thebot, source, "- ACT       Act in channel");	
		notice(thebot, source, "- QUIT      Quit the Bot");	
	        notice(thebot, source, "\2ENF OF HELP\2");
		} else
	                notice(thebot, source, "Sorry you are not my master.");		
 

            } else if (stricmp(cmd, "SAY") == 0) {
                chan = strtok(NULL, " ");
		bmsg = strtok(NULL, "");
                if (!chan || !bmsg) {
                        notice(thebot, source, "Syntax: SAY \2Channel\2 \2Message\2");
                } else {
			if (is_botowner(source, thebot) || is_services_root(source))
			     if (finduser(chan) || findchan(chan))
				send_cmd (thebot, "%s %s :%s", istoken?TOK_PRIVATE:MSG_PRIVATE, chan, bmsg);
			     else
	                        notice(thebot, source, "Can not send to non exist channel or any services bots.");
			else
	                        notice(thebot, source, "Sorry you are not my master.");	
                }

            } else if (stricmp(cmd, "ACT") == 0) {
                chan = strtok(NULL, " ");
		bmsg = strtok(NULL, "");
                if (!chan || !bmsg) {
                        notice(thebot, source, "Syntax: ACT \2Channel\2 \2Message\2");
                } else {
			if (is_botowner(source, thebot) || is_services_root(source))
			     if (finduser(chan) || findchan(chan))
				send_cmd (thebot, "%s %s :ACTION %s",istoken?TOK_PRIVATE:MSG_PRIVATE, chan, bmsg);
			     else
	                        notice(thebot, source, "Can not send to non exist channel or any services bots.");					

			else
	                        notice(thebot, source, "Sorry you are not my master.");		
                }

            } else if (stricmp(cmd, "PART") == 0) {
	        ChannelInfo *ci;
                chan = strtok(NULL, " ");	
		if (!chan)  {
                       	notice(thebot, source, "Syntax: PART \2Channel\2");
			return;
		}

		ci = cs_findchan(chan);
		
       	        if (!chan || !ci) {
                       	notice(thebot, source, "Syntax: PART \2Channel\2");
		} else if (ci->bot && (stricmp(ci->bot,thebot)==0)) {
       	                notice(thebot, source, "Sorry, this bot have been assigned for this channel already.");
                } else {
			if (is_botowner(source, thebot) || is_services_root(source))
        	                send_cmd (thebot, "%s %s", istoken?TOK_PART:MSG_PART, chan);
			else
                	        notice(thebot, source, "Sorry you are not my master.");
                }
            }
	 }
}
static void abotcom(const char *source, ChannelInfo *ci, char *buf, Channel *c)
{
   char *chan = ci->name;
   User *u = finduser (source);
   char *cmd, *param;
   cmd = strtok(buf, " ");
		
   if(u && cmd) {
        /******************** OP Command ********************/

	if ((stricmp(cmd,"!op")==0) || (stricmp(cmd,"!up")==0) || (stricmp(cmd,"`up")==0)) {
	   param = strtok(NULL, " ");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 5)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   } 

	  if (!param) {
		if (!is_on_uchan(u, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (has_chan_status(u, c, CSTATUS_OP))
			notice(ci->bot, source,"You already have op in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "+o", source);

	  } else {
		User *u2 = finduser(param);
		if (!u2 || !is_on_uchan(u2, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if ((ci->flags & CI_OPGUARD) && (get_access(u2, ci) < 5))
			notice(ci->bot, source,"Channel %s set opguard on please respect founder.",chan);
		else if (has_chan_status(u2, c, CSTATUS_OP))
			notice(ci->bot, source,"User %s already have op in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "+o", param);
	  }

	}

       /********************* END OP *************************/
        else if (stricmp(cmd,"!mass")==0) {
		param = strtok(NULL, " ");

		if (!param) {
		        notice (ci->bot, source, ERR_MORE_INFO, s_ChanServ, "MASS");
#ifdef DEFHALFOP
			notice (ci->bot, source,"Syntax: \2MASS\2 \2<+o|-o|+h|-h|+v|-v|kick|msg|notice>\2");
#else
			notice (ci->bot, source,"Syntax: \2MASS\2 \2<+o|-o|+v|-v|kick|msg|notice>\2");
#endif
			return;
		}

		if(!is_on_uchan(u, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (get_access(u, ci) < 10)
		        notice(ci->bot, source,"You don't have access to use mass command.");
		else if (stricmp (param, "-o") == 0) {
		        struct c_userlist *cu, *next;
		        for (cu = c->users; cu; cu = next) {
		            next = cu->next;
			    if (has_chan_status(cu->user, c, CSTATUS_OP))
			            change_cmode (ci->bot, chan, "-o", cu->user->nick);
		        }
			send_cmd(ci->bot, "%s %s :%s has used the \2MassDEOP\2 command on channel %s", istoken?TOK_NOTICE:MSG_NOTICE, chan,source,chan);
			notice (ci->bot, source, "All the OPs (+o) in %s has been removed",chan);
		} else if (stricmp (param, "-v") == 0) {
		        struct c_userlist *cu, *next;
		        for (cu = c->users; cu; cu = next) {
		            next = cu->next;
		            if (has_chan_status(cu->user,c, CSTATUS_VOICE))
		            	change_cmode (ci->bot, chan, "-v", cu->user->nick);
		        }
			send_cmd(ci->bot, "%s %s :%s has used the \2MassDEVOICE\2 command on the channel %s",istoken?TOK_NOTICE:MSG_NOTICE, chan,source,chan);
			notice (ci->bot, source, "All the voices (+v) in %s has been removed",chan);

		} 
		else if (stricmp (param, "kick") == 0) {
		        struct c_userlist *cu, *next_cu;
		        char abuf[BUFSIZE];
			char *reason = strtok (NULL, "");
			snprintf (abuf, sizeof (abuf), "(ChanServ) MassKick by %s", source);
			send_cmd(ci->bot, "%s %s :%s has used the \2MassKICK\2 command on the channel %s",istoken?TOK_NOTICE:MSG_NOTICE,chan,source,chan);
			for (cu = c->users; cu; cu = next_cu)  {
		            next_cu = cu->next;
		            if (stricmp(cu->user->nick, source) !=0) {
		             if(!reason)
				kick_user(ci->bot, chan, cu->user->nick, abuf);
		             else
				kick_user(ci->bot, chan, cu->user->nick, reason);
	               } /* FOR */
	        }
		notice (ci->bot, source, "All the users in %s have been removed",chan);
	    } else if (stricmp (param, "+o") == 0) {
		struct c_userlist *cu, *next;
	        for (cu = c->users; cu; cu = next) {
	            next = cu->next;
        	    if (!has_chan_status(cu->user, c, CSTATUS_OP))
                	change_cmode (ci->bot, chan, "+o", cu->user->nick);
		}
		send_cmd(ci->bot, "%s %s :%s has used the \2MassOP\2 command on the channel %s",istoken?TOK_NOTICE:MSG_NOTICE,chan,source,chan);
		notice (ci->bot, source, "All the users in %s has got the Channel Op (+o) status.",chan);
#ifdef DEFHALFOP
	    }  else if (stricmp (param, "+h") == 0) {
		struct c_userlist *cu, *next;
	        for (cu = c->users; cu; cu = next) {
		    next = cu->next;
	            if (!has_chan_status(cu->user, c, CSTATUS_HALFOP))
	                change_cmode (ci->bot, chan, "+h", cu->user->nick);
        	}
		send_cmd(ci->bot, "%s %s :%s has used the \2MassHOP\2 command on the channel %s",istoken?TOK_NOTICE:MSG_NOTICE,chan,source,chan);
		notice (ci->bot, source, "All the users in %s has got the Channel HOP (+h) status.",chan);
	} else if (stricmp (param, "-h") == 0) {
	        struct c_userlist *cu, *next;
        	for (cu = c->users; cu; cu = next) {
	            next = cu->next;
        	    if (has_chan_status(cu->user, c, CSTATUS_HALFOP))
	               change_cmode (ci->bot, chan, "-h", cu->user->nick);
        	}
		send_cmd(ci->bot, "%s %s :%s has used the \2MassDEOP\2 command on channel %s",istoken?TOK_NOTICE:MSG_NOTICE,chan,source,chan);
		notice (ci->bot, source, "All the OPs (+h) in %s has been removed",chan);
#endif
	} else if (stricmp (param, "+v") == 0) {
	        struct c_userlist *cu, *next;
	        for (cu = c->users; cu; cu = next) {
			next = cu->next;
	            if (!has_chan_status(cu->user,c, CSTATUS_VOICE))
                	change_cmode (ci->bot, chan, "+v", cu->user->nick);
	        }
		send_cmd(ci->bot, "%s %s :%s has used the \2MassVOICE\2 command on the channel %s",istoken?TOK_NOTICE:MSG_NOTICE,chan,source,chan);
		notice (ci->bot, source, "All the users in %s has got the Channel Voice (+v) status.",chan);
	} else if (stricmp (param, "msg") == 0) {
	        struct c_userlist *cu, *next;
		char *extra = strtok (NULL, "");
        	if (!extra) {
	            notice (ci->bot, source,
        	        "Syntax: \2MASS \37%s\37 \37MSG\37 message\2",chan);
	            notice (ci->bot, source, ERR_MORE_INFO, s_ChanServ, "MASS");
	        } else {
	            for (cu = c->users; cu; cu = next) {
	                next = cu->next;
			send_cmd(ci->bot, "%s %s :(MassMSG from %s): %s", istoken?TOK_PRIVATE:MSG_PRIVATE, cu->user->nick,source ,extra);
                    }
		notice(ci->bot, source, "All the users in %s has got your message.",chan);
	        }
	} else if (stricmp (param, "notice") == 0) {
	        struct c_userlist *cu, *next;
		char *extra = strtok (NULL, "");
        	if (!extra) {
	            notice (ci->bot, source,
        	        "Syntax: \2MASS \37%s\37 \37NOTICE\37 message\2",chan);
	            notice (ci->bot, source, ERR_MORE_INFO, s_ChanServ, "MASS");
	        } else {
	            for (cu = c->users; cu; cu = next) {
	                next = cu->next;
			send_cmd(ci->bot, "%s %s :(MassNOTICE from %s): %s",istoken?TOK_NOTICE:MSG_NOTICE,cu->user->nick,source ,extra);
                    }
		notice(ci->bot, source, "All the users in %s has got your notice.",chan);
	        }
       } else {
  	      notice (ci->bot, source,
                "Syntax: \2MASS \37what\37\2");
#ifdef DEFHALFOP
		        notice (ci->bot, source,"Syntax: \2MASS\2 \2<op|deop|halfop|dehalfop|voice|devoice|kick|msg|notice>\2");
#else
		        notice (ci->bot, source,"Syntax: \2MASS\2 \2<op|deop|voice|devoice|kick|msg|notice>\2");
#endif
	}         
   }
       /********************** DEOP *************************/
	else if ((stricmp(cmd,"!deop")==0) || (stricmp(cmd,"!down")==0) || (stricmp(cmd,"`deop")==0)) {
	   param = strtok(NULL, " ");
	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) <5)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_uchan(u, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (!has_chan_status(u, c, CSTATUS_OP))
			notice(ci->bot, source,"You already deop in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "-o", source);

	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (!is_chanop(param, chan))
			notice(ci->bot, source,"User %s already deop in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "-o", param);
	  }
       }
         /*************************** End Deop *********************/
         /*************************** molest ***********************/
	else if (stricmp(cmd,"!molest")==0) {

	if (!is_on_uchan(u, chan))
		notice(ci->bot, source,"You are not in channel %s.", chan);
	else
		dropmode(u, c, ci);

       }
         /*************************** End MoLest *******************/
         /*************************** seen ***********************/
	else if (stricmp(cmd,"!seen")==0) {
	   NickInfo *ni;
	   param = strtok(NULL, " ");
	   if (param) {
		ni = findnick(param);
		if (ni) {
			time_t myseen = CTime - ni->last_seen;
			notice(ci->bot, source,"nick %s last seen %d day(s) %d:%d ago.", param, myseen/86400, (myseen/3600) % 24,(myseen/60) % 60);
		} else {
			notice(ci->bot, source,"nick %s is not registered.", param);
		}
	   }
       }
         /*************************** Voice ************************/
	else if (stricmp(cmd,"!voice")==0) {
	   param = strtok(NULL, " ");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_uchan(u, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (has_chan_status(u, c, CSTATUS_VOICE))
			notice(ci->bot, source,"You already have voice in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "+v", source);

	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (is_voiced(param, chan))
			notice(ci->bot, source,"User %s already have voice in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "+v", param);
	  }
	}
         /*************************** End Voice ********************/
         /*************************** De Voice *********************/
	else if (stricmp(cmd,"!devoice")==0) {
	   param = strtok(NULL, " ");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_uchan(u, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (!has_chan_status(u, c, CSTATUS_VOICE))
			notice(ci->bot, source,"You already devoice in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "-v", source);

	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (!is_voiced(param, chan))
			notice(ci->bot, source,"User %s already devoice in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "-v", param);
	  }
	}
         /*************************** End DeVoice ******************/

         /*************************** Invite *********************/
	else if (stricmp(cmd,"!invite")==0) {
	   param = strtok(NULL, " ");

	   if (!has_chan_status(u, c, CSTATUS_OP) && get_access(u, ci) < 4) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !INVITE <nick>");
	  } else {
		send_cmd(ci->bot, "%s %s %s", istoken?TOK_INVITE:MSG_INVITE, param, chan);
		notice(ci->bot, source,"Invited %s join %s", param, chan);
	  }
	}
         /*************************** End INVITE ******************/

         /*************************** Halfop ***********************/
#ifdef DEFHALFOP
	else if (!stricmp(cmd,"!halfop") || !stricmp(cmd,"!hop")) {
	   param = strtok(NULL, " ");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 5)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_uchan(u, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (has_chan_status(u, c, CSTATUS_HALFOP))
			notice(ci->bot, source,"You already have halfop in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "+h", source);

	  } else {
		User *u2 = finduser(param);
		if (!is_on_uchan(u2, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if ((ci->flags & CI_OPGUARD) && (get_access(u2, ci) < 4))
			notice(ci->bot, source,"Channel %s set opguard on please respect founder.",chan);
		else if (has_chan_status(u2, c, CSTATUS_HALFOP))
			notice(ci->bot, source,"User %s already have halfop in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "+h", param);
	  }
	}
         /*************************** End Halfop *******************/
	else if (!stricmp(cmd,"!dehalfop") || !stricmp(cmd,"!dehop")) {
	   param = strtok(NULL, " ");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 5)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_uchan(u, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (!has_chan_status(u, c, CSTATUS_HALFOP))
			notice(ci->bot, source,"You already dehalfop in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "-h", source);

	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (!is_chanhop(param, chan))
			notice(ci->bot, source,"User %s already dehalfop in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "-h", param);
	  }

	}
#endif
         /*************************** DeHalfop *********************/
         /*************************** Kick *************************/
	else if (stricmp(cmd,"!kick")==0) {
	   char *reason;
	   param = strtok(NULL, " ");
	   reason = strtok(NULL, "");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !KICK <nick> [<reason>]");
	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (
#ifdef DEFHALFOP
is_chanhop(param, chan) || 
#endif
is_chanop(param, chan) 
#ifdef DEFPROTECT
|| is_prots(param, chan)
#endif
)
			notice(ci->bot, source,"I don't kick op or protected user.");
		else {
	               if (!reason)
				kick_user(ci->bot, chan, param, "Requested by an Op");
	               else
				kick_user(ci->bot, chan, param, reason);
                }		
	  }
	}
         /************************** End Kick **********************/
         /************************** Ban ***************************/
	else if (stricmp(cmd,"!ban")==0) {
	   char *btype;
	   param = strtok(NULL, " ");
	   btype = strtok(NULL, " ");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !BAN <nick> ![<type>]");
	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (
#ifdef DEFHALFOP
is_chanhop(param, chan) || 
#endif
is_chanop(param, chan) 
#ifdef DEFPROTECT
|| is_prots(param, chan)
#endif
)
			notice(ci->bot, source,"I don't ban op or protected user.");
		else {
			char mask[BUFSIZE];
		        User *u2 = finduser(param);
			if (u2) {
			if (btype && *btype == '!') {
		               if (!stricmp(btype,"!1")) {
					bcreate_mask(u2,1,mask, sizeof(mask));
	        	        } else {
					bcreate_mask(u2,2,mask, sizeof(mask));
				}
		        } else {
				bcreate_mask(u2,1,mask, sizeof(mask));
			} /* If btype */
			addtban(c, mask, 5, ci->bot);
		      } /* If u2 */
                }		
	  }
	}
         /************************** End Ban ***********************/
         /************************** BKICK *************************/
	else if (stricmp(cmd,"!bkick")==0) {
	   char *reason, *btype;
	   param = strtok(NULL, " ");
	   btype = strtok(NULL, " ");
	   reason = strtok(NULL, "");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !BKICK <nick> ![<type>] [<reason>]");
	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (
#ifdef DEFHALFOP
is_chanhop(param, chan) || 
#endif
is_chanop(param, chan)
#ifdef DEFPROTECT
|| is_prots(param, chan)
#endif
)
			notice(ci->bot, source,"I don't ban kick op or protected user.");
		else {
		       User *u2 = finduser(param);

		       if (btype && *btype == '!') {
				char mask[BUFSIZE];
		               if (!stricmp(btype,"!1")) {
					bcreate_mask(u2,1,mask, sizeof(mask));
					change_cmode(ci->bot, chan, "+b", mask);
		               } else {
					bcreate_mask(u2,2,mask, sizeof(mask));
					change_cmode(ci->bot, chan, "+b", mask);
				}
		       } else {
				char mask[BUFSIZE];
				bcreate_mask(u2,1,mask, sizeof(mask));
				change_cmode(ci->bot, chan, "+b", mask);
			}

/* Come check again */
	               if (!reason) {
		          if (btype && (*btype != '!'))
				kick_user(ci->bot, chan, param, btype);
			  else
				kick_user(ci->bot, chan, param, "Ban kick requested");
		       } else {
		          if (*btype != '!') {
				char *newreason = malloc(strlen(btype) + strlen(reason) + 2);
				snprintf(newreason, strlen(btype) + strlen(reason) + 2,"%s %s", btype, reason);
				kick_user(ci->bot, chan, param, newreason);
				free(newreason);
			   } else
				kick_user(ci->bot, chan, param, reason);
		       }
                }		
	  }
	}
         /************************** Ban Kick **********************/
         /************************** Un Ban ************************/
	else if (stricmp(cmd,"!unban")==0) {
	   char *btype;
	   param = strtok(NULL, " ");
	   btype = strtok(NULL, " ");

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !UNBAN <nick> or <mask>");
	  } else {
	       User *u2 = finduser(param);
	       int i;
	       if (u2) {
		       for (i = c->bancount-1; i > -1; --i) {
		       if (c->bans[i])
		       if (match_usermask2(c->bans[i], u2))
				change_cmode(ci->bot, chan, "-b", c->bans[i]);
		        }
//			deltuban (ci->bot, c, u2);
		} else {
		       for (i = c->bancount-1; i > -1; --i) {
		       if (c->bans[i])
		       if (match_wild_nocase(param, c->bans[i]))
				change_cmode(ci->bot, chan, "-b", c->bans[i]);
		       }
//			delttban (ci->bot, c, param);
		} 
	  }
	}
         /************************** End Unban *********************/

         /************************** HELP **************************/
	else if (stricmp(cmd,"!help")==0) {

	   if (!has_chan_status(u, c, CSTATUS_OP) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

		if (!is_on_uchan(u, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else {
		    const char *hcmd = strtok(NULL, "");
		    char hbuf[BUFSIZE];
		    snprintf(hbuf, sizeof(hbuf), "botcom%s", hcmd ? " " : "");
		    strscpy(hbuf+strlen(hbuf), hcmd ? hcmd : "", sizeof(hbuf)-strlen(hbuf));
		    helpserv(ci->bot, source, hbuf);
		}
       }
         /************************** End Help **********************/
     else if(stricmp(cmd,"!topic")==0) {
         param = strtok(NULL, "");

        if (!ci)
        return;

        if(get_access(u, ci) < 5) {
                notice(ci->bot, source,"You don't have access.");
                return;
        }

        if (ci->flags & CI_TOPICLOCK) {
        if(get_access(u, ci) < ci->topic_allow) {
                notice(ci->bot, source,"You don't have access.");
                return;
        } }
     if (!param) {
        notice(ci->bot, source,"Syntax: \2!TOPIC\2 topic");
     } else if (strlen(param) > 300) {
        notice(ci->bot, source,"Topic too long make it shorter than 300 chars.");
     } else {
        send_cmd(ci->bot, "%s %s :%s (%s)",istoken?TOK_TOPIC:MSG_TOPIC, chan,param,source);
        if (c->topic)
                free(c->topic);
        c->topic = sstrdup(param);
        if (ci->last_topic)
                free(ci->last_topic);
        ci->last_topic = sstrdup(param);
        strscpy(c->topic_setter, ci->last_topic_setter, NICKMAX);
        c->topic_time = ci->last_topic_time;
    }
}
#ifdef GAMES
	else if (!is_tokn(39)) {
	// Disable Game option
	}
	/***********************************************************/
	else if (((stricmp(cmd,"`gnum")==0) || (stricmp(cmd,"!gnum")==0)) && (ci->botflag & CBI_GUESSNUM)){
	   unsigned int theguess;
	   param = strtok(NULL, " ");
	   if (!param) {
		notice(ci->bot, source,"Please Enter 4 digits number.");
		return;
	   } else if (strlen(param) > 4) {
		notice(ci->bot, source,"Please Enter 4 digits number.");
		return;
	   } else {
		theguess = atoi(param);
		if (theguess > c->ngames) {
			send_cmd (ci->bot, "%s %s :ACTION %s \2%d\2 is too big",istoken?TOK_PRIVATE:MSG_PRIVATE,chan,source, theguess);
		} else if (theguess == c->ngames) {
		     double x =  RAND_MAX + 1.0;
		     srand(time(NULL));
		     c->ngames = 1 + rand() * (999/x);
		     u->games++;
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION \37%s\37 \2have Won\2, your score is %d",chan,source, u->games);
		     send_cmd (ci->bot, "NOTICE @%s :Answer is \2%d\2",chan,c->ngames);
	       	send_cmd (ci->bot, "NOTICE %s :New Game start, type \2!GNum <number>\2 to make guess!!",chan);

		} else {
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION \37%s\37 \2%d\2 is too small",chan,source, theguess);
		}
	   } }
       /*****************************************************/
	else if (((stricmp(cmd,"`nextnum")==0) || (stricmp(cmd,"!nextnum")==0)) && (ci->botflag & CBI_GUESSNUM)){
	        double x =  RAND_MAX + 1.0;
		if (!has_chan_status(u, c, CSTATUS_OP)) {
			notice(ci->bot, source,"Only op can do.");
			return;
		}
 	        srand(time(NULL));
	        c->ngames = 1 + rand() * (999/x);
	        send_cmd (ci->bot, "NOTICE @%s :Answer is \2%d\2",chan,c->ngames);
	        send_cmd (ci->bot, "NOTICE %s :New Game start, type \2!GNum <number>\2 to make guess!!",chan);
	   }
       /*****************************************************/
#ifdef WGAME
	else if (((stricmp(cmd,"`gword")==0) || (stricmp(cmd,"!gword")==0)) && (ci->botflag & CBI_GUESSWRD)){
	   param = strtok(NULL, " ");
	   if (!param) {
		notice(ci->bot, source,"Syntax: !GWord <word or char>.");
		return;
	   } else if (strlen(param) > 50) {
		notice(ci->bot, source,"There are no any word longer than 50 chars.");
		return;
	   } else {
		if (!c->wgames || !c->wgame) {
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION Game yet start, to start game one of the op must type \2!nextword\2.", chan);
		} else 	if (stricmp(param, wordgame[c->wgames]) == 0) {
		     char *wordsss;
		     unsigned short i;
		     double x =  RAND_MAX + 1.0;
		     srand(time(NULL));
		     c->wgames = 1 + rand() * (wgameline/x);
		     u->games++;
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION %s \2have Won Guessing Word  Game\2, your Score now is %d",chan,source, u->games);
		     send_cmd (ci->bot, "NOTICE %s :New Word Guessing Game start, type \2!GWord <word or char>\2 to make guess!!",chan);
		     if (c->wgame)
				free(c->wgame);
		     c->wgame = sstrdup(wordgame[c->wgames]);		    
		     wordsss = c->wgame;
		     *wordsss++;
		     for (i=2; i< strlen(wordgame[c->wgames]); i++)
			*wordsss++ = '*';
		     send_cmd (ci->bot, "NOTICE %s :The new word is \2%s\2 Hints: \37%s\37.",chan,c->wgame,wordhelp[c->wgames]);
		     send_cmd (ci->bot, "NOTICE @%s :Answer is \2%s\2",chan,wordgame[c->wgames]);
		     *wordsss = 0;
		} else {
		     char *wordsss, *fakes;
		     unsigned short i;
		     wordsss = c->wgame;
		     fakes = wordgame[c->wgames];
		     for (i=1; i <= strlen(wordgame[c->wgames]); i++) {
			*fakes++;
			*wordsss++;
		        if (toupper(*fakes) == toupper(*param))
		             *wordsss = *param;
		     }
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION The word now is \2%s\2 Hints: \37%s\37.",chan,c->wgame,wordhelp[c->wgames]);
		     *fakes = 0;
		     *wordsss = 0;
		}
	   }
	}
	/*******************************************************/
	else if (((stricmp(cmd,"`nextword")==0) || (stricmp(cmd,"!nextword")==0)) && (ci->botflag & CBI_GUESSWRD)){
	        double x =  RAND_MAX + 1.0;
		char *thisgame;
	        unsigned short i;
		if (
#ifdef DEFHALFOP
has_chan_status(u, c, CSTATUS_HALFOP) || 
#endif
		has_chan_status(u, c, CSTATUS_OP)) {
		     srand(time(NULL));
		     c->wgames = 1 + rand() * (wgameline/x);
		     send_cmd (ci->bot, "NOTICE %s :New Word Guessing Game start, type \2!GWord <word or char>\2 to make guess!!",chan);

		     if (c->wgame)
				free(c->wgame);
		     c->wgame = sstrdup(wordgame[c->wgames]);
		     thisgame = c->wgame;
		     *thisgame++;
		     for (i=2; i< strlen(wordgame[c->wgames]); i++)
			*thisgame++ = '*';
		     send_cmd (ci->bot, "NOTICE %s :The new word is \2%s\2 Hints: \37%s\37.",chan,c->wgame,wordhelp[c->wgames]);
		     send_cmd (ci->bot, "NOTICE @%s :Answer is \2%s\2",chan,wordgame[c->wgames]);
		} else {
			notice(ci->bot, source,"Only op can do.");
			return;
		}   
	   }
#endif /* WGAME */

#endif  /* Games */
   } else {
   }
}
void bmsgchan(const char *source, char *chan, char *buf, User *u)
{
	int cha;
	ChannelInfo *ci;
	char *cmd;
	Channel *c = findchan(chan);
	char *nick;
	int i = 0;
	char *s = buf;
	char morebuf[BUFSIZE];
	UserData *ud;

	if (!c) return;

	nick = sstrdup(source);

	ci = cs_findchan(chan);

	if ((aspychan == 1) && !stricmp(chan,asource))
	     send_cmd(s_OperServ, "%s %s :<%s/%s> %s", istoken?TOK_PRIVATE:MSG_PRIVATE, adest, source, chan, buf);

	if (!ci || !u) goto botend;

	if (ci->flags & CI_CLOSED) goto botend;
	if (ci->flags & CI_FREEZECHAN) goto botend;

	strscpy(morebuf, buf, sizeof(BUFSIZE));
	cmd = strtok(morebuf, " ");


	if (ci && !(ci->bot)) {

		goto botend;

	} else {

		if (!((ci->botflag & CBI_PROTOP) && (has_chan_status(u, c, CSTATUS_VOICE) || has_chan_status(u, c, CSTATUS_OP) 
#ifdef DEFHALFOP
		|| has_chan_status(u, c, CSTATUS_HALFOP) 
#endif
#ifdef DEFPROTECT
		|| has_chan_status(u, c, CSTATUS_PROTECT)
#endif
		))) {
/************************************* REPEAT ************************/
			if (ci->botflag & CBI_KICKREPEAT) {

                        ud = get_user_data(c, u);
                        if (!ud) {
				goto botend;
			}

                        if (ud->lastline && stricmp(ud->lastline, buf)) {
                                free(ud->lastline);
                                ud->lastline = sstrdup(buf);
                                ud->times = 0;
                        } else {
                                if (!ud->lastline) ud->lastline = sstrdup(buf);
                                ud->times++;
                        }
                        if (ud->times >= 4)
                        {
					doonu(u,ci, c, ci->bot, "Don't repeat");
					goto botend;
                        }
			}
		/********** END REPEAT **************/
			if (ci->botflag & CBI_KICKFLOOD) {
				time_t now = CTime;
                        ud = get_user_data(c, u);

                        if (!ud) { goto botend; }

                        if (now - ud->last_start > ci->floodsecs)
                        {
                                ud->last_start = CTime;
                                ud->lines = 0;
                        }

                        ud->lines++;
			if (ud->lines > ci->floodlines)
				ud->lines = 0;

                        if (ud->lines >= ci->floodlines)
                        {
					doonu(u,ci, c, ci->bot, "Don't flood");
					goto botend;
                        }
			}
		/***************** CAP *******************/
			if ((ci->botflag & CBI_KICKCAP) && (cha = strlen(buf) > 10)) {
				do {
		       		if (isallowed(*s))
						if (isupper(*s)) i++;
				} while (*s++);
	
				cha = strlen(buf);

				if (i >= ci->capsmin && i*100/cha>= ci->capspercent)
				{
					doonu(u,ci, c, ci->bot, "Please turn caps lock OFF!");
					goto botend;
				}
			}
		/********** END CAP **************/
                /********** BOLDS ****************/
			if ((ci->botflag & CBI_KICKBOLDS) && strchr(buf, 2)) {
				doonu(u,ci, c, ci->bot, "Don't use bolds on this channel!");
				goto botend;
			}
                /***************** END BOLD **************/
		/***************** COLORS ****************/
			if ((ci->botflag & CBI_KICKCOLORS) && strchr(buf, 3)) {
				doonu(u,ci, c, ci->bot, "Don't use colors on this channel!");
				goto botend;
			}
                /******************** END COLOR ******************/           
		/******************** REVERSES *******************/
			if ((ci->botflag & CBI_KICKREVERSES) && strchr(buf, 22)) {
				doonu(u,ci, c, ci->bot, "Don't use reverses on this channel!");
				goto botend;
			}
		/******************** END REVERSES *******************/
		/******************** UNDERLINES *********************/
			if ((ci->botflag & CBI_KICKUNDERLINES) && strchr(buf, 31)) {
				doonu(u,ci, c, ci->bot, "Don't use underlines on this channel!");
				goto botend;
			}
		/******************* END UNDERLINES ******************/
			if ((ci->botflag & CBI_KICKBADWORDS) && is_chanbadwords(ci, buf)) {
				doonu(u,ci, c, ci->bot, "Don't use bad word on this channel!!");
				goto botend;
			}
		}	
		/******************** BOT COMMAND ****************/
		if (ci->botflag & CBI_BOTCOM) {
			abotcom(source, ci, buf, c);
			goto botend;
		}
	}
botend:
	free(nick);
}
/************************************************************************/
void botkick(User *u2, ChannelInfo *ci, Channel *c, char *buf)
{
	if (is_on_uchan(u2, ci->name) && ci->bot) {
		char mask[BUFSIZE];
		bcreate_mask(u2, 1, mask, sizeof(mask));
		addtban(c, mask, 2, ci->bot);
		kick_user(ci->bot, ci->name, u2->nick, buf);
	}
}
/************************************************************************/
void dropmode(User *u, Channel *c, ChannelInfo *ci)
{
	if (is_on_uchan(u, ci->name)) {
		if (has_chan_status(u, c, CSTATUS_OP))
		change_cmode (ci->bot, ci->name, "-o", u->nick);
#ifdef DEFPROTECT
		if (has_chan_status(u, c, CSTATUS_PROTECT))
		change_cmode (ci->bot, ci->name, "-a", u->nick);
#endif
#ifdef DEFHALFOP
		if (has_chan_status(u, c, CSTATUS_HALFOP))
		change_cmode (ci->bot, ci->name, "-h", u->nick);
#endif
		if (has_chan_status(u, c, CSTATUS_VOICE))
		change_cmode (ci->bot, ci->name, "-v", u->nick);
	}		
}
static void doonu(User *u, ChannelInfo *ci, Channel *c, char *abot, char *areason) {

	if (!is_uoper(u)) {
	if (
#ifdef DEFHALFOP
	has_chan_status(u, c, CSTATUS_HALFOP) || 
#endif
	has_chan_status(u, c, CSTATUS_OP)) {
		notice(abot, u->nick,areason);
		dropmode(u, c, ci);
	} else {
		botkick(u, ci, c, areason);
	}
	}
}
